import assert from 'node:assert';
import { callCbOnQuery, callCbOnQueryError, callCbOnQueryResp, callCbOnStart, } from './event.js';
export function builderBindEvents(kmore, refTable, caseConvert, ctx, kmoreQueryId) {
    assert(caseConvert, 'caseConvert must be defined');
    const queryCtxOpts = {
        wrapIdentifierCaseConvert: kmore.wrapIdentifierCaseConvert,
        postProcessResponseCaseConvert: caseConvert,
        kmoreQueryId,
        columns: [],
    };
    const opts = {
        ctx,
        dbId: kmore.dbId,
        cbs: kmore.eventCallbacks,
        kmoreQueryId,
    };
    const refTable2 = refTable
        .queryContext(queryCtxOpts)
        .on('start', (builder) => {
        const columns2 = pickColumnsFromBuilder(builder);
        if (columns2.length) {
            columns2.forEach(row => queryCtxOpts.columns.push(row));
        }
        return callCbOnStart({
            ...opts,
            builder,
        });
    })
        .on('query', (data) => callCbOnQuery({
        ...opts,
        data,
    }))
        .on('query-response', (resp, respRaw) => callCbOnQueryResp({
        ...opts,
        _resp: resp,
        respRaw,
    }))
        .on('query-error', async (err, data) => {
        const trx = kmore.getTrxByKmoreQueryId(kmoreQueryId);
        await kmore.finishTransaction(trx);
        return callCbOnQueryError({
            ...opts,
            err,
            data,
        });
    });
    // .on('error', (ex: unknown) => {
    //   void ex
    // })
    // .on('end', () => {
    //   void 0
    // })
    return refTable2;
}
function pickColumnsFromBuilder(builder) {
    const ret = [];
    // @ts-ignore
    const statements = builder._statements;
    if (Array.isArray(statements) && statements.length) {
        statements.forEach((statement) => {
            if (statement.grouping !== 'columns') {
                return;
            }
            const { value } = statement;
            if (Array.isArray(value)) {
                value.forEach((item) => {
                    if (typeof item === 'string') {
                        ret.push({ [item]: item });
                    }
                    else if (typeof item === 'object') {
                        ret.push(item);
                    }
                });
            }
        });
    }
    return ret;
}
//# sourceMappingURL=builder.event.js.map