import { CaseConvertTable, CaseType, DbScopedColsByKey, DbScopedColsByTableType, JoinTableWithCaseConvert, SplitScopedColumn, StrKey, UnwrapArrayMember } from '@waiting/shared-types';
import { DbDict } from 'kmore-types';
import type { Knex } from 'knex';
import { PropagationType } from './propagation.types.js';
/**
 * - 0: No paging
 * - 1: Paging, PagingMeta on response Array
 * - 2: paging, wrap response as `PageWrapType`
 */
type PagingCategory = 0 | 1 | 2;
export type KmoreQueryBuilder<D extends {} = {}, CaseConvert extends CaseType = CaseType, EnablePage extends PagingCategory = 0, TRecord extends {} = any, TResult = any> = QueryBuilderExtName<D> & QueryBuilderExtMethod<D, CaseConvert, EnablePage, TRecord> & QueryBuilder<D, CaseConvert, EnablePage, TRecord, AddPagingMeta<TResult, EnablePage>>;
type OmitQueryBuilderKeys = 'select' | 'where' | 'orderBy' | 'columns' | keyof Knex.ChainableInterface;
interface QueryBuilder<D extends {} = {}, CaseConvert extends CaseType = CaseType, EnablePage extends PagingCategory = 0, TRecord extends {} = any, TResult = any> extends Knex.ChainableInterface<AddPagingMeta<ResolveResult<TResult>, EnablePage>>, Omit<Knex.QueryBuilder<TRecord, TResult>, OmitQueryBuilderKeys> {
    select: Select<D, CaseConvert, EnablePage, TRecord, TResult>;
    where: Where<D, CaseConvert, EnablePage, TRecord, TResult>;
    orderBy: OrderBy<D, CaseConvert, EnablePage, TRecord, TResult>;
    columns: Select<D, CaseConvert, EnablePage, TRecord, TResult>;
}
export type DbQueryBuilder<Context, D, Prefix extends string, CaseConvert extends CaseType> = {
    [tb in keyof D as `${Prefix}${tb & string}`]: TbQueryBuilder<D, CaseConvert, CaseConvertTable<D[tb], CaseConvert>, Context>;
};
export type TbQueryBuilder<D extends {}, CaseConvert extends CaseType, TRecord extends {}, Context> = (options?: Partial<TbQueryBuilderOptions<Context>>) => KmoreQueryBuilder<D, CaseConvert, 0, TRecord, TRecord[]>;
export interface TbQueryBuilderOptions<Context> {
    ctx: Context;
    ctxBuilderPreProcessor: CtxBuilderPreProcessor | undefined;
    ctxBuilderResultPreProcessor: CtxBuilderResultPreProcessor | undefined;
    ctxExceptionHandler: CtxExceptionHandler | undefined;
}
export declare enum QueryBuilderExtKey {
    caseConvert = "caseConvert",
    kmoreQueryId = "kmoreQueryId",
    dbDict = "dbDict",
    dbId = "dbId",
    tablesJoin = "_tablesJoin",
    pagingType = "pagingType",
    transactionalProcessed = "transactionalProcessed",
    trxPropagateOptions = "trxPropagateOptions",
    trxPropagated = "trxPropagated",
    rowLockLevel = "rowLockLevel"
}
interface QueryBuilderExtName<D extends {} = {}> {
    caseConvert: CaseType;
    kmoreQueryId: symbol;
    dbDict: DbDict<D>;
    dbId: string;
    _tablesJoin: string[];
    pagingType?: 'counter' | 'pager';
    trxPropagateOptions?: TrxPropagateOptions;
    trxPropagated?: boolean;
    /**
     * Propagation rowlock level
     * @default {@link RowLockLevel}
     */
    rowLockLevel: RowLockLevel | undefined;
    transactionalProcessed: boolean | undefined;
}
export interface TrxPropagateOptions {
    entryKey: string;
    key: string;
    dbId: string;
    type: PropagationType;
    path: string;
    className: string;
    funcName: string;
    methodName: string;
    line: number;
    column: number;
    /**
     * @default {@link RowLockLevel.ForShare}
     */
    readRowLockLevel: RowLockLevel;
    /**
     * @default {@link RowLockLevel.ForUpdate}
     */
    writeRowLockLevel: RowLockLevel;
}
/**
 * Used for `@Transactional()` decorator
 */
export declare enum RowLockLevel {
    ForShare = "FOR_SHARE",
    ForUpdate = "FOR_UPDATE",
    None = "None"
}
interface QueryBuilderExtMethod<D extends {} = {}, CaseConvert extends CaseType = CaseType, EnablePage extends PagingCategory = 0, TRecord extends {} = any> {
    smartCrossJoin: SmartJoin<D, CaseConvert, EnablePage, TRecord>;
    smartInnerJoin: SmartJoin<D, CaseConvert, EnablePage, TRecord>;
    smartJoin: SmartJoin<D, CaseConvert, EnablePage, TRecord>;
    smartLeftJoin: SmartJoin<D, CaseConvert, EnablePage, TRecord>;
    smartRightJoin: SmartJoin<D, CaseConvert, EnablePage, TRecord>;
    autoPaging: AutoPaging<D, CaseConvert, TRecord>;
}
type AutoPaging<D extends {} = {}, CaseConvert extends CaseType = CaseType, TRecord extends {} = any> = <Wrap extends boolean | undefined = false>(options?: Partial<PagingOptions>, wrapOutput?: Wrap) => KmoreQueryBuilder<D, CaseConvert, CalcPagingCat<Wrap>, TRecord, TRecord[]>;
type CalcPagingCat<T> = T extends true ? 2 : 1;
type AddPagingMeta<TSelection, EnablePage extends PagingCategory = 0> = EnablePage extends 0 ? TSelection : TSelection extends (infer R)[] ? EnablePage extends 2 ? WrapPageOutput<R> : PageRawType<R> : TSelection;
type WrapPageOutput<T> = T extends PageWrapType ? T : PageWrapType<T>;
type SmartJoin<D extends {} = {}, CaseConvert extends CaseType = CaseType, EnablePage extends PagingCategory = 0, TResult = unknown[]> = <TRecord1 = UnwrapArrayMember<TResult>, C2 extends DbScopedColsByKey<D> = DbScopedColsByKey<D>, C1 extends DbScopedColsByKey<D> = DbScopedColsByTableType<D, TRecord1>, TTable2 extends StrKey<D> = SplitScopedColumn<D, C2>[0], TRecord2 extends D[TTable2] = D[TTable2], TResult2 = JoinTableWithCaseConvert<TRecord1, TRecord2 extends any ? D[TTable2] : TRecord2, TTable2, CaseConvert>>(
/**
 * scoped column name, e.g. 'tb_name.col_name',
 * <tb_name> is the table name be joined, <col_name> is the column name
 */
scopedColumnBeJoined: C2, 
/**
 * scoped column name, e.g. 'tb_name.col_name',
 * <tb_name> is the upstream table name , <col_name> is the column name
 */
scopedColumn: C1 extends C2 ? never : C1) => KmoreQueryBuilder<D, CaseConvert, EnablePage, TResult2, TResult2[]>;
export interface PagingOptions {
    /**
     * @default true
     */
    enable: boolean;
    /**
     * Current page number, start from 1
     * @default 1
     */
    page: number;
    /**
     * @default 10
     */
    pageSize: number;
}
/**
 * Note: keyof PagingMeta is not enumerable
 */
export type PageRawType<T = unknown> = T[] & PagingMeta;
export interface PageWrapType<T = unknown> extends PagingMeta {
    rows: T[];
}
export interface PagingMeta {
    /**
     * Total count of rows in table
     *
     * @note This is the number of query response rows,
     *  not the number of rows in the current page,
     *  also not the number of pages.
     */
    total: number;
    /**
     * Current page number, start from 1
     */
    page: number;
    /**
     * Number of rows of each page.
     * The number rows of the last page may be less than this value
     */
    pageSize: number;
}
interface Select<D extends {} = {}, CaseConvert extends CaseType = CaseType, EnablePage extends PagingCategory = 0, TRecord extends {} = any, TResult = unknown[]> extends AliasQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult>, ColumnNameQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult> {
    (): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult>;
    <TResult2 = ArrayIfAlready<TResult, any>, TInnerRecord extends {} = any, TInnerResult = any>(...subQueryBuilders: readonly KmoreQueryBuilder<D, CaseConvert, EnablePage, TInnerRecord, TInnerResult>[]): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult2>;
    <TResult2 = ArrayIfAlready<TResult, any>, TInnerRecord extends {} = any, TInnerResult = any>(subQueryBuilders: readonly KmoreQueryBuilder<D, CaseConvert, EnablePage, TInnerRecord, TInnerResult>[]): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult2>;
}
interface AliasQueryBuilder<D extends {} = {}, CaseConvert extends CaseType = CaseType, EnablePage extends PagingCategory = 0, TRecord extends {} = any, TResult = unknown[]> {
    <AliasUT extends Knex.InferrableColumnDescriptor<Knex.ResolveTableType<TRecord>>[], TResult2 = ArrayIfAlready<TResult, DeferredKeySelectionNS.Augment<UnwrapArrayMember<TResult>, Knex.ResolveTableType<TRecord>, IncompatibleToAlt<ArrayMember<AliasUT>, string, never>, Knex.IntersectAliases<AliasUT>>>>(...aliases: AliasUT): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult2>;
    <AliasUT extends Knex.InferrableColumnDescriptor<Knex.ResolveTableType<TRecord>>[], TResult2 = ArrayIfAlready<TResult, DeferredKeySelectionNS.Augment<UnwrapArrayMember<TResult>, Knex.ResolveTableType<TRecord>, IncompatibleToAlt<ArrayMember<AliasUT>, string, never>, Knex.IntersectAliases<AliasUT>>>>(aliases: AliasUT): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult2>;
    <AliasUT extends (Dict | string)[], TResult2 = ArrayIfAlready<TResult, DeferredKeySelectionNS.Augment<UnwrapArrayMember<TResult>, Knex.ResolveTableType<TRecord>, IncompatibleToAlt<ArrayMember<AliasUT>, string, never>, Knex.IntersectAliases<AliasUT>>>>(...aliases: AliasUT): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult2>;
    <AliasUT extends (Dict | string)[], TResult2 = ArrayIfAlready<TResult, DeferredKeySelectionNS.Augment<UnwrapArrayMember<TResult>, TRecord, IncompatibleToAlt<ArrayMember<AliasUT>, string, never>, Knex.IntersectAliases<AliasUT>>>>(aliases: AliasUT): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult2>;
}
interface ColumnNameQueryBuilder<D extends {} = {}, CaseConvert extends CaseType = CaseType, EnablePage extends PagingCategory = 0, TRecord extends {} = any, TResult = unknown[]> {
    (columnName: '*'): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, ArrayIfAlready<TResult, DeferredKeySelection<TRecord, string>>>;
    <ColNameUT extends keyof Knex.ResolveTableType<TRecord>, TResult2 = DeferredKeySelectionNS.Augment<UnwrapArrayMember<TResult>, Knex.ResolveTableType<TRecord>, ColNameUT & string>[]>(...columnNames: readonly ColNameUT[]): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult2>;
    <ColNameUT extends keyof Knex.ResolveTableType<TRecord>, TResult2 = DeferredKeySelectionNS.Augment<UnwrapArrayMember<TResult>, Knex.ResolveTableType<TRecord>, ColNameUT & string>[]>(columnNames: readonly ColNameUT[]): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult2>;
    <TResult2 = DeferredKeySelectionNS.Augment<UnwrapArrayMember<TResult>, SafePartial<TRecord>, keyof TRecord & string>[]>(...columnNames: readonly Knex.ColumnDescriptor<TRecord, TResult>[]): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult2>;
    <TResult2 = DeferredKeySelectionNS.Augment<UnwrapArrayMember<TResult>, SafePartial<TRecord>, keyof TRecord & string>[]>(columnNames: readonly Knex.ColumnDescriptor<TRecord, TResult>[]): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult2>;
}
interface OrderBy<D extends {} = {}, CaseConvert extends CaseType = CaseType, EnablePage extends PagingCategory = 0, TRecord extends {} = any, TResult = unknown[]> {
    (columnName: keyof TRecord | QueryBuilder, order?: 'asc' | 'desc' | undefined, nulls?: 'first' | 'last' | undefined): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult>;
    (columnName: string | QueryBuilder, order?: string | undefined, nulls?: string | undefined): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult>;
    (columnDefs: (keyof TRecord | Readonly<{
        column: keyof TRecord | QueryBuilder;
        order?: 'asc' | 'desc' | undefined;
        nulls?: 'first' | 'last' | undefined;
    }>)[]): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult>;
    (columnDefs: (string | Readonly<{
        column: string | QueryBuilder;
        order?: string | undefined;
        nulls?: string | undefined;
    }>)[]): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult>;
}
type ArrayIfAlready<T1, T2> = AnyToUnknown<T1> extends any[] ? T2[] : T2;
type ArrayMember<T> = T extends (infer M)[] ? M : never;
type AnyToUnknown<T> = unknown extends T ? unknown : T;
type CurlyCurlyToAny<T> = T extends unknown ? (<U>() => U extends T ? 0 : 1) extends <U>() => U extends {} ? 0 : 1 ? any : T : never;
type Dict<T = any> = Record<string, T>;
type IncompatibleToAlt<T, TBase, TAlt> = T extends TBase ? T : TAlt;
type MappedAliasType<TBase, TAliasMapping> = {} & {
    [K in keyof TAliasMapping]: TAliasMapping[K] extends keyof TBase ? TBase[TAliasMapping[K]] : any;
};
type UnknownOrCurlyCurlyToAny<T> = [UnknownToAny<T> | CurlyCurlyToAny<T>][0];
type UnknownToAny<T> = unknown extends T ? any : T;
type AugmentParams<TTarget, TParams> = TParams extends {} ? keyof TParams extends never ? TTarget : {} & TTarget & TParams : TTarget;
type PartialOrAny<TBase, TKeys> = Boxed<TKeys> extends Boxed<never> ? {} : Boxed<TKeys> extends Boxed<keyof TBase> ? SafePick<TBase, TKeys & keyof TBase> : any;
interface Boxed<T> {
    _value: T;
}
type SafePick<T, K extends keyof T> = T extends {} ? Pick<T, K> : any;
type SafePartial<T> = Partial<AnyOrUnknownToOther<T, {}>>;
type AnyOrUnknownToOther<T1, T2> = unknown extends T1 ? T2 : T1;
type Any = DeferredKeySelection<any, any, any, any, any, any, any>;
interface DeferredKeySelection<TBase, TKeys extends string, THasSelect extends true | false = false, TAliasMapping extends {} = {}, TSingle extends boolean = false, TIntersectProps extends {} = {}, TUnionProps = never> {
    _base: TBase;
    _hasSelection: THasSelect;
    _keys: TKeys;
    _aliases: TAliasMapping;
    _single: TSingle;
    _intersectProps: TIntersectProps;
    _unionProps: TUnionProps;
}
export declare namespace DeferredKeySelectionNS {
    type SetBase<TSelection, TBase> = TSelection extends DeferredKeySelection<any, infer TKeys, infer THasSelect, infer TAliasMapping, infer TSingle, infer TIntersectProps, infer TUnionProps> ? DeferredKeySelection<TBase, TKeys, THasSelect, TAliasMapping, TSingle, TIntersectProps, TUnionProps> : DeferredKeySelection<TBase, never>;
    type ReplaceBase<TSelection, TBase> = UnwrapArrayMember<TSelection> extends Any ? ArrayIfAlready<TSelection, SetBase<UnwrapArrayMember<TSelection>, TBase>> : unknown extends UnwrapArrayMember<TSelection> ? ArrayIfAlready<TSelection, SetBase<unknown, TBase>> : TSelection;
    type SetSingle<TSelection, TSingle extends boolean> = TSelection extends DeferredKeySelection<infer TBase, infer TKeys, infer THasSelect, infer TAliasMapping, any, infer TIntersectProps, infer TUnionProps> ? DeferredKeySelection<TBase, TKeys, THasSelect, TAliasMapping, TSingle, TIntersectProps, TUnionProps> : never;
    type AddKey<TSelection, TKey extends string> = TSelection extends DeferredKeySelection<infer TBase, infer TKeys, any, infer TAliasMapping, infer TSingle, infer TIntersectProps, infer TUnionProps> ? DeferredKeySelection<TBase, TKeys | TKey, true, TAliasMapping, TSingle, TIntersectProps, TUnionProps> : DeferredKeySelection<unknown, TKey, true>;
    type AddAliases<TSelection, T extends {}> = TSelection extends DeferredKeySelection<infer TBase, infer TKeys, infer THasSelect, infer TAliasMapping, infer TSingle, infer TIntersectProps, infer TUnionProps> ? DeferredKeySelection<TBase, TKeys, THasSelect, TAliasMapping & T, TSingle, TIntersectProps, TUnionProps> : DeferredKeySelection<unknown, never, false, T>;
    type AddUnionMember<TSelection, T> = TSelection extends DeferredKeySelection<infer TBase, infer TKeys, infer THasSelect, infer TAliasMapping, infer TSingle, infer TIntersectProps, infer TUnionProps> ? DeferredKeySelection<TBase, TKeys, THasSelect, TAliasMapping, TSingle, TIntersectProps, TUnionProps | T> : DeferredKeySelection<TSelection, never, false, {}, false, {}, T>;
    type Augment<T, TBase, TKey extends string, TAliasMapping extends {} = {}> = AddAliases<AddKey<SetBase<T, TBase>, TKey>, TAliasMapping>;
    type ResolveOne<TSelection> = TSelection extends DeferredKeySelection<infer TBase, infer TKeys, infer THasSelect, infer TAliasMapping, infer TSingle, infer TIntersectProps, infer TUnionProps> ? UnknownOrCurlyCurlyToAny<AugmentParams<AnyToUnknown<TBase> extends {} ? TSingle extends true ? TKeys extends keyof TBase ? TBase[TKeys] : any : AugmentParams<true extends THasSelect ? PartialOrAny<TBase, TKeys> : TBase, MappedAliasType<TBase, TAliasMapping>> : unknown, TIntersectProps> | TUnionProps> : TSelection;
    type Resolve<TSelection> = TSelection extends Any ? Knex.ResolveTableType<ResolveOne<TSelection>> : TSelection extends Any[] ? Knex.ResolveTableType<ResolveOne<TSelection[0]>>[] : TSelection extends (infer I)[] ? UnknownOrCurlyCurlyToAny<Knex.ResolveTableType<I>>[] : UnknownOrCurlyCurlyToAny<Knex.ResolveTableType<TSelection>>;
}
type ResolveResult<S, EnablePage extends PagingCategory = 0> = AddPagingMeta<DeferredKeySelectionNS.Resolve<S>, EnablePage>;
type ComparisonOperator = '=' | '>' | '>=' | '<' | '<=' | '<>';
export interface Where<D extends {} = {}, CaseConvert extends CaseType = CaseType, EnablePage extends PagingCategory = 0, TRecord extends {} = any, TResult = any> extends WhereRaw<D, CaseConvert, EnablePage, TRecord, TResult> {
    (raw: Knex.Raw): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult>;
    (callback: Knex.QueryCallback<TRecord, TResult>): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult>;
    (object: Knex.DbRecord<Knex.ResolveTableType<TRecord>>): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult>;
    (object: Readonly<Object>): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult>;
    <T extends keyof Knex.ResolveTableType<TRecord>>(columnName: T, value: Knex.DbColumn<Knex.ResolveTableType<TRecord>[T]> | null): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult>;
    (columnName: string, value: Knex.Value | null): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult>;
    <T extends keyof Knex.ResolveTableType<TRecord>>(columnName: T, operator: ComparisonOperator, value: Knex.DbColumn<Knex.ResolveTableType<TRecord>[T]> | null): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult>;
    (columnName: string, operator: string, value: Knex.Value | null): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult>;
    <T extends keyof Knex.ResolveTableType<TRecord>, TRecordInner extends {}, TResultInner>(columnName: T, operator: ComparisonOperator, value: Knex.QueryBuilder<TRecordInner, TResultInner>): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult>;
    <TRecordInner extends {}, TResultInner>(columnName: string, operator: string, value: Knex.QueryBuilder<TRecordInner, TResultInner>): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult>;
    (left: Knex.Raw, operator: string, right: Knex.Value | null): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult>;
    <TRecordInner extends {}, TResultInner>(left: Knex.Raw, operator: string, right: Knex.QueryBuilder<TRecordInner, TResultInner>): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult>;
}
interface WhereRaw<D extends {} = {}, CaseConvert extends CaseType = CaseType, EnablePage extends PagingCategory = 0, TRecord extends {} = any, TResult = unknown[]> extends Knex.RawQueryBuilder<TRecord, TResult> {
    (condition: boolean): KmoreQueryBuilder<D, CaseConvert, EnablePage, TRecord, TResult>;
}
export type CtxBuilderPreProcessor = (builder: KmoreQueryBuilder) => Promise<{
    builder: KmoreQueryBuilder;
}>;
export type CtxBuilderResultPreProcessor<T = unknown> = (options: CtxBuilderResultPreProcessorOptions<T>) => Promise<T>;
export type CtxExceptionHandler = (options: CtxExceptionHandlerOptions) => Promise<never>;
export interface CtxBuilderResultPreProcessorOptions<Resp = unknown> {
    kmoreQueryId: symbol;
    kmoreTrxId: symbol | undefined;
    response: Resp;
    transactionalProcessed: boolean | undefined;
    trxPropagateOptions: TrxPropagateOptions | undefined;
    trxPropagated: boolean | undefined;
    /**
     * Propagation rowlock level
     * @default {@link RowLockLevel}
     */
    rowLockLevel: RowLockLevel | undefined;
}
export interface CtxExceptionHandlerOptions extends Omit<CtxBuilderResultPreProcessorOptions, 'response'> {
    exception: unknown;
}
export {};
//# sourceMappingURL=builder.types.d.ts.map