import { RecordCamelKeys, RecordPascalKeys, RecordSnakeKeys } from '@waiting/shared-types';
import type { Knex } from 'knex';
import { CaseType, KnexConfig, QueryContext } from './types.js';
export declare function getCurrentTime(dbh: Knex, clientType: KnexConfig['client']): Promise<string>;
/**
 * Convert identifier (field) to snakecase
 */
export declare function wrapIdentifier(value: string, origImpl: (input: string) => string, queryContext?: QueryContext): string;
export declare function postProcessResponse<T extends PostProcessInput = PostProcessInput>(result: T, queryContext?: QueryContext): T | PostProcessRespRet<T, QueryContext['postProcessResponseCaseConvert']>;
export type PostProcessPlain = number | string | undefined | null | boolean;
export type PostProcessRecord = Record<string, PostProcessPlain> | object;
export type PostProcessArray = PostProcessPlain[];
export type PostProcessInput = PostProcessPlain | PostProcessRecord | PostProcessArray;
export type PostProcessRespRet<T extends PostProcessInput, CaseConvert extends CaseType | undefined> = T extends PostProcessPlain ? T : T extends PostProcessArray ? PostProcessArray : T extends PostProcessRecord ? PostProcessRecordCaseConvert<T, CaseConvert> : never;
type PostProcessRecordCaseConvert<T extends PostProcessRecord, CaseConvert extends CaseType | undefined> = CaseConvert extends CaseType.camel ? RecordCamelKeys<T> : CaseConvert extends CaseType.pascal ? RecordPascalKeys<T> : CaseConvert extends CaseType.snake ? RecordSnakeKeys<T> : T;
/**
 * Convert keys of result to camelcase, if input is object
 */
export declare function postProcessResponseToCamel<T extends PostProcessInput = PostProcessInput>(result: T, _queryContext?: QueryContext): PostProcessRespRet<T, CaseType.camel>;
/**
 * Convert keys of result to PascalCase, if input is object
 */
/**
 * Convert keys of result to snake_case, if input is object
 */
export declare function postProcessResponseToSnake<T extends PostProcessInput = PostProcessInput>(result: T, _queryContext?: QueryContext): PostProcessRespRet<T, CaseType.snake>;
export declare function genCamelKeysFrom<From extends PostProcessRecord>(input: From): RecordCamelKeys<From, '_'>;
export declare function genSnakeKeysFrom<From extends PostProcessRecord>(input: From): RecordSnakeKeys<From, '_'>;
/**
 * @description only one level
 */
export declare function mergeDoWithInitData<T extends Record<string, unknown> | object>(initDoData: T, input?: Record<string, unknown> | object): T;
export {};
//# sourceMappingURL=helper.d.ts.map