/* eslint-disable @typescript-eslint/no-redundant-type-constituents */
/* eslint-disable @typescript-eslint/no-explicit-any */
import assert from 'assert';
import { defaultPropDescriptor } from './config.js';
import { KmoreProxyKey } from './types.js';
export async function processThenRet(options) {
    const { ctxBuilderResultPreProcessor, ctxExceptionHandler, input, kmore, kmoreQueryId, kmoreTrxId, rowLockLevel, transactionalProcessed, trxPropagated, trxPropagateOptions, done, reject, } = options;
    const pm1 = ctxBuilderResultPreProcessor
        ? input.then((response) => {
            return ctxBuilderResultPreProcessor({
                kmoreQueryId,
                kmoreTrxId,
                response,
                transactionalProcessed,
                trxPropagateOptions,
                trxPropagated,
                rowLockLevel,
            });
        })
        : input;
    const pm2 = ctxExceptionHandler
        ? pm1.catch((ex) => ctxExceptionHandler({
            kmoreQueryId,
            kmoreTrxId,
            transactionalProcessed,
            trxPropagateOptions,
            trxPropagated,
            rowLockLevel,
            exception: ex,
        }))
        : pm1;
    return pm2
        .catch((ex) => processTrxOnEx(kmore, kmoreQueryId, ex))
        // eslint-disable-next-line @typescript-eslint/no-unsafe-return
        .then((resp) => processThen(resp, done))
        .catch((ex) => processEx({
        ex,
        // kmore,
        // kmoreQueryId: target.kmoreQueryId,
        reject,
    }))
        .then((resp) => {
        if (resp && typeof resp === 'object'
            && !Object.hasOwn(resp, KmoreProxyKey.getThenProxyProcessed)) {
            Object.defineProperty(resp, KmoreProxyKey.getThenProxyProcessed, {
                ...defaultPropDescriptor,
                enumerable: false,
                writable: true,
                value: true,
            });
        }
        return resp;
    });
}
function processThen(resp, cb) {
    if (resp && typeof resp === 'object'
        && !Object.hasOwn(resp, KmoreProxyKey.getThenProxyProcessed)) {
        Object.defineProperty(resp, KmoreProxyKey.getThenProxyProcessed, {
            ...defaultPropDescriptor,
            enumerable: false,
            writable: true,
            value: true,
        });
    }
    return typeof cb === 'function' ? cb(resp) : resp;
}
async function processEx(options) {
    const { ex, reject } = options;
    // await processTrxOnEx(kmore, kmoreQueryId)
    const ex2 = ex instanceof Error
        ? ex
        : typeof ex === 'string'
            ? new Error(ex)
            : new Error('Kmore Error when executing then()', { cause: ex });
    if (typeof reject === 'function') {
        return reject(ex2);
    }
    return Promise.reject(ex2);
}
async function processTrxOnEx(kmore, kmoreQueryId, ex) {
    assert(kmoreQueryId, 'kmoreQueryId should be set on QueryBuilder');
    console.error('processTrxOnEx', kmoreQueryId, ex);
    const trx = kmore.getTrxByKmoreQueryId(kmoreQueryId);
    if (trx) { // also processed on event `query-error`
        await kmore.finishTransaction(trx).catch(console.error);
    }
    if (ex) {
        return Promise.reject(ex);
    }
}
//# sourceMappingURL=proxy.get.helper.js.map